/******************************** SBB **********************************
**
** Project      :       Multiple-Threaded Library (MTL)
**
** Author       :       Zhang Xiangyang
**
** Date         :       November 4, 1996
**
** Version      :       1.0
**              :       Copyright (C) 1996, Zhang Xiangyang
**                
** File name    :       pthread.c
**
** Description  :
**
**      This file is part of pthread Library, which provides the thread 
**      interface compatible with POSIX 1003.4a Reference.
**
**      Report problems and direct all questions to:
**       zhangx@sebb.bel.alcatel.be
**
********************************HISTORY*********************************
**   DATE    INITIALS	COMMENTS		
**   ----    --------	--------
** 04/11/96  XZHANG     Creation for version 1 only
**
************************************************************************/

#ifndef _PTHREAD_MUTEX_H_INCLUDED

    
/*
 * Macros to lock and unlock a mutex
 * precondition: 	SET_KERNEL_FLAG
 * postcondition: 	CLEAR_KERNEL_FLAG
 */


#define mac_mutex_lock(mutex, p)\
{\
  int   not_done = 1;\
\
  do\
  {\
    if (!mutex->lock)\
    {\
      mutex->lock = TRUE;\
      mutex->owner = p;\
      tcb[p].mutex = mutex;\
      not_done = 0;\
    }\
    else\
    {\
      mac_mutex_wait(mutex, p);\
      CONTEXT_SWITCH(TCB_WAITING);\    
    }\
  } while (not_done);\
}


#define mac_mutex_unlock(mutex)\
{\
  int       num;\
\
  if (!mutex->lock)\
  {\
    fprintf(stderr,"unlock err.\n");\
  }\
  else\
  {\
    mutex->owner = NO_OWNER;\
    mutex->lock = FALSE;\
    mac_mutex_wakeup(mutex, &num);\
    mac_thread_ready(&num);\
	if (num > 0) \
      CONTEXT_SWITCH(TCB_READY);\
  }\
}

#define _PTHREAD_MUTEX_H_INCLUDED
#endif /* !_PTHREAD_MUTEX_H_INCLUDED */
