/*
 * This file is used to find REQUEST prime numbers.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#define 	WORKERS         8
#define		REQUEST         100

#define		check(status, string)\
			if (status == -1) perror(string)

pthread_mutex_t		prime_list,
					current_mutex,
					cond_mutex;
pthread_cond_t		cond_var;
int					current_num = -1,
					thread_hold = 1,
					count = 0,
					primes[REQUEST];
pthread_t			threads[WORKERS];


/* Work thread routine */

int prime_search(pthread_key_t arg)
{
	int			numerator,
				denominator,
				cut_off,
				notifiee,
				prime,
				my_number,
				status,
				not_done = 1;
				
	my_number = (int) arg;
	
	status = pthread_mutex_lock(&cond_mutex);
	check(status, "1 : Mutex_lock bad status.\n");
	while (thread_hold)
	{
		status = pthread_cond_wait(&cond_var, &cond_mutex);
		check(status, "2 : cond_wait bad status.\n");
	}

	status = pthread_mutex_unlock(&cond_mutex);
	check(status, "3 : Mutex_unlock bad status.\n");
	
	while (not_done)
	{
		pthread_testcancel();

		status = pthread_mutex_lock(&current_mutex);
		check(status, "4 : Mutex_lock bad status.\n");

		current_num += 2;  /* skip even number */
		numerator = current_num;
		
		status = pthread_mutex_unlock(&current_mutex);
		check(status, "5 : Mutex_unlock bad status.\n");
		
		cut_off = numerator/2 + 1;
		prime = 1;
		
		for (denominator = 2; ((denominator < cut_off) && (prime)); 
								denominator++)
			prime = numerator % denominator;
		
		if (prime != 0)
		{
			pthread_setcancel(CANCEL_OFF);
			
			status = pthread_mutex_lock(&prime_list);
			check(status, "6 : Mutex_lock bad status.\n");

			if (count < REQUEST)
			{
				primes[count] = numerator;
				count ++;
			}
			else if (count == REQUEST)
			{
				not_done = 0;
				count ++;
				for (notifiee = 0; notifiee < WORKERS; ++notifiee)
				{
					if (notifiee != my_number)
					{
						pthread_cancel(&threads[notifiee]);
						check(status, "7 : cancel bad status.\n");
					}
				}
			}
			else
			{
				not_done = 0;
				count++;
				printf("count = %d\n", count);
			}

			status = pthread_mutex_unlock(&prime_list);
			check(status, "8 : Mutex_unlock bad status.\n");

			pthread_setcancel(CANCEL_ON);
		}
		pthread_testcancel();
	}
	pthread_exit(my_number);
}

main()
{
	int 	work_num,
			exit_value,
			list,
			status, 
			index1, index2,
			temp,
			not_done;

	pthread_init();
	status = pthread_mutex_init(&prime_list, NULL);
	check(status, "9 : mutex_init bad status.\n");
	status = pthread_mutex_init(&cond_mutex, NULL);
	check(status, "10 : mutex_init bad status.\n");
	status = pthread_mutex_init(&current_mutex, NULL);
	check(status, "11 : mutex_init bad status.\n");

	status = pthread_cond_init(&cond_var, NULL);
	check(status, "12 : cond_init bad status.\n");
	
	for (work_num = 0; work_num < WORKERS; ++ work_num)
	{
		status = pthread_create(&threads[work_num],
								NULL,
								prime_search,
								work_num+1);
		check(status, "13 : create bad status.\n");
	}

	status = pthread_mutex_lock(&cond_mutex);
	check(status, "14 : Mutex_lock bad status.\n");

	thread_hold = 0;
	status = pthread_cond_broadcast(&cond_var);

	status = pthread_mutex_unlock(&cond_mutex);
	check(status, "15 : Mutex_unlock bad status.\n");
	
	for (work_num = 0; work_num < WORKERS; ++ work_num)
	{
		status = pthread_join(&threads[work_num],
								&exit_value);
		check(status, "16 : join bad status.\n");

		if (exit_value == work_num+1)
			printf("Thread terminated normally\n");

		status = pthread_detach(&threads[work_num]);
		check(status, "17 : detach bad status.\n");
	}
	
	printf("The list of %d primes follows:\n", REQUEST);
	for (list = 0; list < REQUEST; ++list)
		printf("prime[%d] \t= %d\n", list, primes[list]);
	printf("\n\n");

	not_done = 1;
	for (index1 = 1; ((index1 < REQUEST) && (not_done)); index1++)
		for (index2 = 0; index2 < index1; index2++)
		{
			if (primes[index1] < primes[index2])
			{
				temp = primes[index2];
				primes[index2] = primes[index1];
				primes[index1] = temp;
				not_done = 0;
			}
		}
	
	printf("The list of %d primes follows:\n", REQUEST);
	for (list = 0; list < REQUEST; ++list)
		printf("prime[%d] \t= %d\n", list, primes[list]);
	printf("\n\n");
	return (0);
}